DEFAULT_MODEL_PHASE = "default"
PHASE_INTENTION = "intention"
PHASE_QUERY_REWRITE = "query_rewrite"
PHASE_SEARCH_SUMMARY = "search_summary"
PHASE_SUMMARY = "summary"
PHASE_ORGANIZE_REFERENCE = "organize_reference"

PHASE_SEARCH_INTENTION = "search_intention"
PHASE_SEARCH_INTENTION_BASIC = "search_intention_basic"
PHASE_SEARCH_REWRITE = "search_query_rewrite"
PHASE_SEARCH = "search"
PHASE_RAG_INTENTION = "rag_intention"
PHASE_RAG_REWRITE = "rag_query_rewrite"
PHASE_RAG_RETRIEVE = "rag_retrieve"

CUSTOM_VARIABLE_KEY_TASK_DESCRIPTION = "task_description"
CUSTOM_VARIABLE_KEY_KNOWLEDGE_BASE_DESCRIPTION = "knowledge_base_description"

ACTION_CONFIG_KEY_ACTION_NAME = "action_name"
ACTION_CONFIG_KEY_SOURCE_TYPE = "source_type"
ACTION_NAME_RAG_SNAKE = "rag"
ACTION_NAME_WEBBROWSING_SNAKE = "web_browsing"

DEFAULT_RESULT_MAPPING = {"需要": True, "不需要": False}

# v2 plugins
CALCULATOR_V2_ACTION_NAME = "Calculator"
CALCULATOR_V2_TOOL_NAME = "Calculator"
LINK_READER_V2_ACTION_NAME = "LinkReader"
LINK_READER_V2_TOOL_NAME = "LinkReader"
CONTENT_PLUGIN_ACTION_NAME = "content_plugin"
CONTENT_PLUGIN_TOOL_NAME = "search"
